# Generated date columns.
DATE_YEAR_COLUMN = 'year'
DATE_MONTH_COLUMN = 'month'
DATE_WEEK_COLUMN = 'week'

# Generated columns available for --group-by.
# Use sorted() for reproducible output.
GROUP_BY_GENERATED_COLUMNS = {
    DATE_YEAR_COLUMN,
    DATE_MONTH_COLUMN,
    DATE_WEEK_COLUMN,
}
